#[repr(C, usize)]pub enum CodeIteratorResult<'p> {
NullHandlePointer,
Ok {
handle: Box<CodeIteratorHandle<'p>>,
},
Err(OwnedBytes),
}Expand description
A result type returned from FFI functions that create an code hash iterator
Variants
NullHandlePointer
The caller provided a null pointer to a proof handle.
Ok
Building the iterator was successful and the iterator handle is returned
Fields

handle: Box<CodeIteratorHandle<'p>>An opaque pointer to the CodeIteratorHandle.
The value should be freed with fwd_code_hash_iter_free
Err(OwnedBytes)
An error occurred and the message is returned as an OwnedBytes.
The caller must call fwd_free_owned_bytes to free the memory
associated with this error.