pub struct HashKey(/* private fields */);Expand description
A database hash key, used in FFI functions that require hashes. This type requires no allocation and can be copied freely and dropped without any additional overhead.
This is useful because it is the same size as 4 words which is equivalent to 2 heap-allocated slices (pointer + length each), or 1.5 vectors (which uses an extra word for allocation capacity) and it can be passed around without needing to allocate or deallocate memory.