#[repr(C)]pub struct NextKeyRange {
pub start_key: OwnedBytes,
pub end_key: Maybe<OwnedBytes>,
}Expand description
A key range that should be fetched to continue iterating through a range
or change proof that was truncated. Represents a half-open range
[start_key, end_key). If end_key is None, the range is unbounded
and continues to the end of the keyspace.
Fields
start_key: OwnedBytesThe start key of the next range to fetch.
end_key: Maybe<OwnedBytes>If set, a non-inclusive upper bound for the next range to fetch. If not set, the range is unbounded (this is the final range).